
class TestSession:
    def test_can_call(self):
        from tdameritrade import session
        session.TDASession()

    def test_init_header(self):
        from tdameritrade.session import TDASession
        token = ""
        TDASession().set_token(token)



import os
import pandas as pd
from .session import TDASession
from .urls import ACCOUNTS, INSTRUMENTS, QUOTES, SEARCH, HISTORY, OPTIONCHAIN, MOVERS
from .exceptions import handle_error_response


def response_is_valid(resp):
    valid_codes = [200, 201]
    return resp.status_code in valid_codes


class TDClient(object):
    def __init__(self, access_token=None, accountIds=None):
        self._token = access_token or os.environ['ACCESS_TOKEN']
        self.accountIds = accountIds or []
        self.session = TDASession()
        if self._token:
            self.session.set_token(self._token)

    def _headers(self):
        return {'Authorization': 'Bearer ' + self._token}

    def _request(self, method, params=None, *args, **kwargs):
        resp = self.session.request('GET', method, params=params, *args, **kwargs)
        if not response_is_valid(resp):
            handle_error_response(resp)

        return resp

    # TODO: output results to self.accountIds
    def accounts(self, positions=False, orders=False):
        ret = {}

        if positions or orders:
            fields = '?fields='
            if positions:
                fields += 'positions'
                if orders:
                    fields += ',orders'
            elif orders:
                fields += 'orders'
        else:
            fields = ''

        if self.accountIds:
            for acc in self.accountIds:
                resp = self._request(ACCOUNTS + str(acc) + fields, headers=self._headers())
                ret[acc] = resp.json()

        else:
            resp = self._request(ACCOUNTS + fields, headers=self._headers())
            for account in resp.json():
                ret[account['securitiesAccount']['accountId']] = account

        return ret

    def accountsDF(self):
        return pd.json_normalize(self.accounts())

    def transactions(self, acc=None, type=None, symbol=None, start_date=None, end_date=None):
        if acc is None:
            acc = self.accounts
        transactions = ACCOUNTS + str(acc) + "/transactions"
        resp = self._request(transactions,
                             headers=self._headers(),
                             params={
                                 'type': type,
                                 'symbol': symbol,
                                 'startDate': start_date,
                                 'endDate': end_date
                             }).json()

        return resp

    def transactionsDF(self, acc, **kwargs):
        return pd.json_normalize(self.transactions(acc, kwargs))

    def search(self, symbol, projection='symbol-search'):
        resp = self._request(SEARCH,
                             headers=self._headers(),
                             params={'symbol': symbol,
                                     'projection': projection}).json()
        return resp

    def searchDF(self, symbol, projection='symbol-search'):
        ret = []
        dat = self.search(symbol, projection)
        for symbol in dat:
            ret.append(dat[symbol])
        return pd.DataFrame(ret)

    def fundamental(self, symbol):
        return self.search(symbol, 'fundamental')

    def fundamentalDF(self, symbol):
        return self.searchDF(symbol, 'fundamental')

    def instrument(self, cusip):
        resp = self._request(INSTRUMENTS + str(cusip),
                             headers=self._headers()).json()
        return resp

    def instrumentDF(self, cusip):
        return pd.DataFrame(self.instrument(cusip))

    def quote(self, symbol):
        resp = self._request(QUOTES,
                             headers=self._headers(),
                             params={'symbol': symbol.upper()}).json()
        return resp

    def quoteDF(self, symbol):
        x = self.quote(symbol)
        return pd.DataFrame(x).T.reset_index(drop=True)

    def history(self, symbol, **kwargs):
        resp = self._request(HISTORY % symbol,
                             headers=self._headers(),
                             params=kwargs).json()
        return resp

    def historyDF(self, symbol, **kwargs):
        x = self.history(symbol, **kwargs)
        df = pd.DataFrame(x['candles'])
        df['datetime'] = pd.to_datetime(df['datetime'], unit='ms')
        return df

    def options(self, symbol):
        resp = self._request(OPTIONCHAIN,
                             headers=self._headers(),
                             params={'symbol': symbol.upper()}).json()
        return resp

    def optionsDF(self, symbol):
        ret = []
        dat = self.options(symbol)
        for date in dat['callExpDateMap']:
            for strike in dat['callExpDateMap'][date]:
                ret.extend(dat['callExpDateMap'][date][strike])
        for date in dat['putExpDateMap']:
            for strike in dat['putExpDateMap'][date]:
                ret.extend(dat['putExpDateMap'][date][strike])

        df = pd.DataFrame(ret)
        for col in ('tradeTimeInLong', 'quoteTimeInLong', 'expirationDate', 'lastTradingDay'):
            df[col] = pd.to_datetime(df[col], unit='ms')
        return df

    def movers(self, index, direction='up', change_type='percent'):
        resp = self._request(MOVERS % index,
                             headers=self._headers(),
                             params={'direction': direction,
                                     'change_type': change_type}).json()
        return resp

    def saved_orders(self, account_id, json_order):
        saved_orders = ACCOUNTS + account_id + "/savedorders"
        resp = self._request(saved_orders,
                             headers=self._headers(),
                             json=json_order).json()
        return resp

    def orders(self, account_id, json_order):
        orders = ACCOUNTS + account_id + "/orders"
        resp = self._request(orders,
                             headers=self._headers(),
                             json=json_order
                             ).json()
        return resp



import os
import os.path
import sys
import requests
import time
from selenium import webdriver
from shutil import which
import urllib.parse as up


def authentication(client_id, redirect_uri, tdauser=None, tdapass=None):
    client_id = client_id + '@AMER.OAUTHAP'
    url = 'https://auth.tdameritrade.com/auth?response_type=code&redirect_uri=' + up.quote(redirect_uri) + '&client_id=' + up.quote(client_id)

    options = webdriver.ChromeOptions()

    if sys.platform == 'darwin':
        # MacOS
        if os.path.exists("/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"):
            options.binary_location = "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"
        elif os.path.exists("/Applications/Chrome.app/Contents/MacOS/Google Chrome"):
            options.binary_location = "/Applications/Chrome.app/Contents/MacOS/Google Chrome"
    elif 'linux' in sys.platform:
        # Linux
        options.binary_location = which('google-chrome') or which('chrome') or which('chromium')

    else:
        # Windows
        if os.path.exists('C:/Program Files (x86)/Google/Chrome/Application/chrome.exe'):
            options.binary_location = 'C:/Program Files (x86)/Google/Chrome/Application/chrome.exe'
        elif os.path.exists('C:/Program Files/Google/Chrome/Application/chrome.exe'):
            options.binary_location = 'C:/Program Files/Google/Chrome/Application/chrome.exe'

    chrome_driver_binary = which('chromedriver') or "/usr/local/bin/chromedriver"
    driver = webdriver.Chrome(chrome_driver_binary, chrome_options=options)

    driver.get(url)

    # Set tdauser and tdapass from environemnt if TDAUSER and TDAPASS environment variables were defined
    tdauser = tdauser or os.environ.get('TDAUSER', '')
    tdapass = tdapass or os.environ.get('TDAPASS', '')

    # Fully automated oauth2 authentication (if tdauser and tdapass were intputed into the function, or found as environment variables)
    if tdauser and tdapass:
        ubox = driver.find_element_by_id('username')
        pbox = driver.find_element_by_id('password')
        ubox.send_keys(tdauser)
        pbox.send_keys(tdapass)
        driver.find_element_by_id('accept').click()

        driver.find_element_by_id('accept').click()
        while 1:
            try:
                code = up.unquote(driver.current_url.split('code=')[1])
                if code != '':
                    break
                else:
                    time.sleep(2)
            except (TypeError, IndexError):
                pass
    else:
        input('after giving access, hit enter to continue')
        code = up.unquote(driver.current_url.split('code=')[1])

    driver.close()

    resp = requests.post('https://api.tdameritrade.com/v1/oauth2/token',
                         headers={'Content-Type': 'application/x-www-form-urlencoded'},
                         data={'grant_type': 'authorization_code',
                               'refresh_token': '',
                               'access_type': 'offline',
                               'code': code,
                               'client_id': client_id,
                               'redirect_uri': redirect_uri})
    if resp.status_code != 200:
        raise Exception('Could not authenticate!')
    return resp.json()


def refresh_token(refresh_token, client_id):
    resp = requests.post('https://api.tdameritrade.com/v1/oauth2/token',
                         headers={'Content-Type': 'application/x-www-form-urlencoded'},
                         data={'grant_type': 'refresh_token',
                               'refresh_token': refresh_token,
                               'client_id': client_id})
    if resp.status_code != 200:
        raise Exception('Could not authenticate!')
    return resp.json()


def main():
    client_id = input('client id:')
    redirect_uri = input('redirect uri:')
    print(authentication(client_id, redirect_uri))




def handle_error_response(resp):
    codes = {
        400: ValidationError,
        401: InvalidAuthToken,
        500: ServerError,
        403: Forbidden,
        404: NotFound,
        -1: TDAAPIError
    }

    raise codes[resp.status_code]()


class TDAAPIError(Exception):
    response = None
    data = {}
    code = -1
    message = "An unknown error occurred"

    def __init__(self, message=None, code=None, data={}, response=None):
        self.response = response
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data

    def __str__(self):
        if self.code:
            return '{}: {}'.format(self.code, self.message)
        return self.data


class ValidationError(TDAAPIError):
    pass


class InvalidAuthToken(TDAAPIError):
    pass


class ServerError(TDAAPIError):
    pass


class Forbidden(TDAAPIError):
    pass


class NotFound(TDAAPIError):
    pass



# for Coverage
from mock import patch, MagicMock
import pytest



@pytest.fixture
def tdclient():
    from tdameritrade import TDClient
    tdc = TDClient('test', [1, 2])
    return tdc


class TestExtension:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_init(self, tdclient):
        tdclient._headers()

    def test_request(self, tdclient):
        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = [MagicMock()]
            tdclient._request("http://goodurl.com", None)

    def test_request_exception(self, tdclient):
        from tdameritrade.exceptions import InvalidAuthToken
        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 401
            # resp = [{"status_code": 401}]
            # #m.return_value = MagicMock()
            # #m.return_value.status_code = {"status_code": 401}
            # mock = MagicMock()
            # mock.return_value = resp
            # m.return_value = mock
            # #m.return_value.json.return_value = [MagicMock()]

            with pytest.raises(InvalidAuthToken):
                tdclient._request("http://goodurl.com", None)

    def test_accounts(self, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:

            m.return_value.status_code = 200
            m.return_value.json.return_value = [MagicMock()]
            tdclient.accounts()

            m.return_value.status_code = 200
            m.return_value.json.return_value = [MagicMock()]

            m.return_value.status_code = 200
            m.return_value.json.return_value = [{'test': 1, 'test2': 2}]
            tdclient.accountsDF()

    def test_search(self, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = {'aapl': {'test': 1, 'test2': 2}}
            tdclient.search('aapl')
            tdclient.searchDF('aapl')

    def test_instrument(self, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = {'aapl': {'test': 1, 'test2': 2}}
            tdclient.instrument('aapl')
            tdclient.instrumentDF('aapl')

    def test_quote(self, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = {'aapl': {'test': 1, 'test2': 2}}
            tdclient.quote('aapl')
            tdclient.quoteDF('aapl')

    def test_history(self, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = {'candles': [{'datetime': 1, 'test2': 2}]}
            tdclient.history('aapl')
            tdclient.historyDF('aapl')

    def test_movers(self):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 200
            m.return_value.json.return_value = {'aapl': {'test'}}

    @pytest.fixture
    def json_order(cls):
        from tdameritrade.tests.JSONS import TEST_BUY_MARKET_STOCK

    def test_orders(self, json_order, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 201
            m.return_value.json.return_value = [MagicMock()]
            tdclient.orders('1234567', json_order)

    def test_saved_orders(self, json_order, tdclient):

        with patch('tdameritrade.session.TDASession.request') as m:
            m.return_value.status_code = 201
            m.return_value.json.return_value = [MagicMock()]
            tdclient.saved_orders('1234567', json_order)




TEST_BUY_MARKET_STOCK = '''
{
  "orderType": "MARKET",
  "session": "NORMAL",
  "duration": "DAY",
  "orderStrategyType": "SINGLE",
  "orderLegCollection": [
    {
      "instruction": "Buy",
      "quantity": 15,
      "instrument": {
        "symbol": "XYZ",
        "assetType": "EQUITY"
      }
    }
  ]
}
'''


import pytest
from mock import MagicMock

class TestExceptions:

    def test_can_call(self):
        from tdameritrade.exceptions import TDAAPIError
        TDAAPIError()

    def test_handle_error_response(self):
        from tdameritrade.exceptions import handle_error_response, InvalidAuthToken
        resp = MagicMock(status_code=401)

        with pytest.raises(InvalidAuthToken):
            handle_error_response(resp)


from setuptools import setup, find_packages
from codecs import open
import io
import os
import os.path

pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))
name = 'tdameritrade'


def get_version(file, name='__version__'):
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]

version = get_version(pjoin(here, name, '_version.py'))

with open(pjoin(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

requires = [
    'ipython>=7.0.1',
    'pillow>=5.3.0',
    'pandas>=0.22.0',
    'ujson>=1.35',
]

requires_dev = [
    'flake8>=3.7.8',
    'mock',
    'pytest>=4.3.0',
    'pytest-cov>=2.6.1',
    'Sphinx>=1.8.4',
    'sphinx-markdown-builder>=0.5.2',
] + requires

setup(
    name=name,
    version=version,
    description='APIs for TD Ameritrade',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/timkpaine/{name}'.format(name=name),
    author='Tim Paine',
    author_email='t.paine154@gmail.com',
    license='Apache 2.0',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    keywords='finance data',
    zip_safe=False,
    packages=find_packages(exclude=[]),
    entry_points={
        'console_scripts': ['tdameritrade-auth = tdameritrade.auth:main']
    },
    install_requires=requires,
    extras_require={
        'dev': requires_dev,
    },
)




BASE = 'https://api.tdameritrade.com/v1/'
ACCOUNTS = BASE + 'accounts/'
SEARCH = BASE + 'instruments'
INSTRUMENTS = SEARCH + '/'
QUOTES = BASE + 'marketdata/quotes'
HISTORY = BASE + 'marketdata/%s/pricehistory'
OPTIONCHAIN = BASE + 'marketdata/chains'
MOVERS = BASE + 'marketdata/%s/movers'
